import re
from typing import Optional

from abstra_internals.entities.forms.widgets.response_types import PhoneResponse
from abstra_internals.entities.forms.widgets.widget_base import InputWidget


class PhoneInput(InputWidget):
    type = "phone-input"
    value: PhoneResponse

    def __init__(
        self,
        label: str,
        *,
        key: Optional[str] = None,
        placeholder: Optional[str] = None,
        required: bool = True,
        hint: Optional[str] = None,
        full_width: bool = False,
        disabled: bool = False,
        value: Optional[PhoneResponse] = None,
        errors: Optional[str] = None,
        invalid_message: Optional[str] = None,
    ):
        self.label = label
        self._key = key or label
        self.placeholder = placeholder
        self.required = required
        self.hint = hint
        self.full_width = full_width
        self.disabled = disabled
        self.value = value or self.make_empty()
        self.errors = self._init_errors(errors)
        self.invalid_message = invalid_message

    def make_empty(self):
        return PhoneResponse(raw="", masked="", country_code="", national_number="")

    def render(self):
        return {
            "type": self.type,
            "key": self._key,
            "label": self.label,
            "value": self.serialize_value(),
            "placeholder": self.placeholder,
            "required": self.required,
            "hint": self.hint,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
            "errors": self.errors,
            "invalidMessage": self.invalid_message,
        }

    def _run_validators(self):
        errors = []
        if self.required and (
            self.value.country_code == "" or self.value.national_number == ""
        ):
            errors.append("i18n_error_required_field")
        return errors

    def is_value_unset(self):
        return self.value.raw == ""

    def set_value(self, value, set_errors=False):
        if isinstance(value, PhoneResponse):
            self.value = value
        elif isinstance(value, dict):
            self.value = self.parse_value(value)
        elif value is None:
            self.value = self.make_empty()
        if set_errors:
            self.set_errors()

    def serialize_value(self):
        if isinstance(self.value, dict):
            return {
                "countryCode": re.sub(
                    "\\D", "", str(self.value.get("country_code", ""))
                ),
                "nationalNumber": re.sub(
                    "\\D", "", str(self.value.get("national_number", ""))
                ),
            }
        if isinstance(self.value, PhoneResponse):
            return {
                "countryCode": re.sub("\\D", "", str(self.value.country_code)),
                "nationalNumber": re.sub("\\D", "", str(self.value.national_number)),
            }
        return {"countryCode": "", "nationalNumber": ""}

    def parse_value(self, value) -> PhoneResponse:
        if value is None:
            return self.make_empty()
        masked = f"+{value.get('countryCode')} {value.get('nationalNumber')}"
        raw = re.sub("\\D", "", masked)
        return PhoneResponse(
            raw=raw,
            masked=masked,
            country_code=f"+{value.get('countryCode')}",
            national_number=value.get("nationalNumber"),
        )



from abstra_internals.controllers.sdk_context import SDKContextStore


class ReuseCacheInterface:
    def __init__(self) -> None:
        self.controller = SDKContextStore.get_by_thread().form_sdk
        self.seq = 0

    def reuse(self, func, *args, **kwargs):
        self.controller.reuse(func, *args, **kwargs)


def reuse(func, *args, **kwargs):
    ReuseCacheInterface().reuse(func, *args, **kwargs)



from typing import List, Optional, Union

from abstra_internals.entities.forms.widgets.widget_base import InputWidget


class PasswordInput(InputWidget):
    type = "password-input"
    value: str

    def __init__(
        self,
        label: str,
        *,
        key: Optional[str] = None,
        hint: Optional[str] = None,
        required: bool = True,
        full_width: bool = False,
        lowercase_required: bool = False,
        uppercase_required: bool = False,
        special_required: bool = False,
        digit_required: bool = False,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        size: Optional[int] = None,
        pattern: Optional[str] = None,
        autocomplete: Optional[str] = None,
        placeholder: Optional[str] = None,
        disabled: bool = False,
        secret: bool = False,
        errors: Optional[Union[List[str], str]] = None,
        value: str = "",
    ):
        self.label = label
        self._key = key or label
        self.value = value
        self.hint = hint
        self.required = required
        self.full_width = full_width
        self.lowercase_required = lowercase_required
        self.uppercase_required = uppercase_required
        self.special_required = special_required
        self.digit_required = digit_required
        self.min_length = min_length
        self.max_length = max_length
        self.size = size
        self.pattern = pattern
        self.autocomplete = autocomplete
        self.placeholder = placeholder
        self.disabled = disabled
        self.secret = secret
        self.errors = self._init_errors(errors)
        self.value = value

    def render(self):
        return {
            "value": self.value,
            "type": self.type,
            "key": self._key,
            "label": self.label,
            "hint": self.hint,
            "required": self.required,
            "fullWidth": self.full_width,
            "lowercaseRequired": self.lowercase_required,
            "uppercaseRequired": self.uppercase_required,
            "specialRequired": self.special_required,
            "digitRequired": self.digit_required,
            "minLength": self.min_length,
            "maxLength": self.max_length,
            "size": self.size,
            "pattern": self.pattern,
            "autocomplete": self.autocomplete,
            "placeholder": self.placeholder,
            "disabled": self.disabled,
            "secret": self.secret,
            "errors": self.errors,
        }

    def parse_value(self, value: str) -> str:
        return value



from typing import Any, Callable, Optional

from abstra_internals.entities.forms.widgets.widget_base import InputWidget


class CustomInput(InputWidget):
    type = "custom-input"
    value: Any

    def __init__(
        self,
        html_body: str,
        *,
        key: Optional[str] = None,
        label: str = "",
        html_head: str = "",
        height: int = 200,
        css: str = "",
        js: str = "",
        value: Any = None,
        full_width: bool = False,
        change_event: Optional[Callable] = None,
    ):
        self.html_body = html_body
        self._key = key or str(hash(html_body))
        self.label = label
        self.html_head = html_head
        self.height = height
        self.css = css
        self.js = js
        self.value = value
        self.full_width = full_width
        self.change_event = change_event

    def render(self):
        return {
            "type": self.type,
            "key": self._key,
            "label": self.label,
            "value": self.value,
            "htmlBody": self.html_body,
            "height": self.height,
            "htmlHead": self.html_head,
            "css": self.css,
            "js": self.js,
            "fullWidth": self.full_width,
            "errors": self.errors,
        }

    def parse_value(self, value) -> Any:
        if self.change_event is not None:
            self.value = self.change_event(value)
        else:
            self.value = value
        return self.value



import io
from typing import Union

from abstra_internals.entities.forms.widgets.file_upload import upload_widget_file
from abstra_internals.entities.forms.widgets.widget_base import OutputWidget


class FileOutput(OutputWidget):
    type = "file-output"

    def __init__(
        self,
        file: Union[str, io.IOBase],
        *,
        download_text: str = "Download",
        full_width: bool = False,
    ):
        self.file = upload_widget_file(file)
        self.download_text = download_text
        self.full_width = full_width

    def render(self):
        return {
            "type": self.type,
            "fileUrl": self.file,
            "downloadText": self.download_text,
            "fullWidth": self.full_width,
        }



import re
from typing import List, Optional, Union

from abstra_internals.entities.forms.widgets.widget_base import InputWidget


class EmailInput(InputWidget):
    type = "email-input"
    value: str

    def __init__(
        self,
        label: str,
        *,
        key: Optional[str] = None,
        required: bool = True,
        hint: Optional[str] = None,
        placeholder: Optional[str] = "",
        full_width: bool = False,
        invalid_email_message: str = "i18n_error_invalid_email",
        disabled: bool = False,
        errors: Optional[Union[List[str], str]] = None,
        value: str = "",
    ):
        self.label = label
        self._key = key or label
        self.required = required
        self.hint = hint
        self.placeholder = placeholder
        self.full_width = full_width
        self.invalid_email_message = invalid_email_message
        self.disabled = disabled
        self.value = value
        self.errors = self._init_errors(errors)

    def render(self):
        return {
            "type": self.type,
            "key": self._key,
            "label": self.label,
            "value": self.value,
            "placeholder": self.placeholder,
            "required": self.required,
            "hint": self.hint,
            "fullWidth": self.full_width,
            "invalidEmailMessage": self.invalid_email_message,
            "disabled": self.disabled,
            "errors": self.errors,
        }

    def _run_validators(self) -> List[str]:
        errors = super()._run_validators()
        if len(self.value) == 0:
            return errors
        is_valid_email = bool(
            re.search(
                '^(([^<>()[\\]\\\\.,;:\\s@"]+(\\.[^<>()[\\]\\\\.,;:\\s@"]+)*)|(".+"))@((\\[[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\])|(([a-zA-Z\\-0-9]+\\.)+[a-zA-Z]{2,}))$',
                self.value.strip().lower(),
            )
        )
        if not is_valid_email:
            errors.append(self.invalid_email_message)
        return errors



import json
from typing import TYPE_CHECKING, Any, List, Optional, Union

from abstra_internals.entities.forms.widgets.widget_base import (
    InputWidget,
    MultipleHandler,
)

if TYPE_CHECKING:
    import pandas as pd


class PandasRowSelectionInput(InputWidget):
    type = "pandas-row-selection-input"
    value: Union[List[Any], Any] = None

    def __init__(
        self,
        df: "pd.DataFrame",
        *,
        key: Optional[str] = None,
        required: bool = True,
        hint: Optional[str] = None,
        full_width: bool = True,
        display_index: bool = False,
        disabled: bool = False,
        label: str = "",
        filterable: bool = False,
        multiple: bool = False,
        initial_value: Optional[Union[List, Any]] = None,
        min: Optional[int] = None,
        max: Optional[int] = None,
        page_size: int = 10,
        errors: Optional[Union[List[str], str]] = None,
        value: Optional[Union[List[Any], Any]] = None,
    ):
        self.df = df
        self._key = key or label
        self.required = required
        self.hint = hint
        self.full_width = full_width
        self.display_index = display_index
        self.disabled = disabled
        self.label = label
        self.filterable = filterable
        self.multiple = multiple
        self.empty_value = [] if self.multiple else None
        self.value = initial_value or self.empty_value
        self.min = min
        self.max = max
        self.multiple_handler = MultipleHandler(
            self.multiple, self.min, self.max, self.required
        )
        self.page_size = page_size
        self.errors = self._init_errors(errors)
        self.value = value

    def serialize_table(self):
        if self.df is None:
            import pandas as pd

            df = pd.DataFrame({"change the": [1, 2, 3], "df property": [4, 5, 6]})
            df_json = df.to_json(orient="table")
            if not isinstance(df_json, str):
                raise Exception("df.to_json() did not return a string")
            serialized = json.loads(df_json)
            del serialized["schema"]["pandas_version"]
            return serialized
        df_json = str(self.df.to_json(orient="table"))
        serialized = json.loads(df_json)
        del serialized["schema"]["pandas_version"]
        return serialized

    def _run_validators(self) -> List[str]:
        return self.multiple_handler.validate(self.value)

    def serialize_value(self) -> List:
        return self.multiple_handler.value_to_list(self.value)

    def parse_value(self, value):
        return self.multiple_handler.value_to_list_or_value(value)

    def render(self):
        return {
            "type": self.type,
            "key": self._key,
            "hint": self.hint,
            "table": self.serialize_table(),
            "required": self.required,
            "fullWidth": self.full_width,
            "displayIndex": self.display_index,
            "disabled": self.disabled,
            "label": self.label,
            "multiple": self.multiple,
            "filterable": self.filterable,
            "value": self.serialize_value(),
            "errors": self.errors,
            "min": self.min,
            "max": self.max,
            "pageSize": self.page_size,
        }



import copy
from dataclasses import dataclass
from typing import Callable, Dict, Generator, List, Optional, Tuple, TypedDict, Union

from abstra_internals.entities.forms.form_state import State
from abstra_internals.entities.forms.widgets.widget_base import BaseWidget, InputWidget


@dataclass
class Button:
    label: str


class NextButton(Button):
    def __init__(self):
        super().__init__("i18n_next_action")


class BackButton(Button):
    def __init__(self):
        super().__init__("i18n_back_action")


Template = List[BaseWidget]
TemplateWithButtons = Tuple[Template, Optional[List[Button]]]

TemplateFunction = Callable[[State], Optional[Union[Template, TemplateWithButtons]]]
TemplateGenerator = Generator[Union[Template, TemplateWithButtons], None, None]
TemplateGeneratorFunction = Callable[[State], TemplateGenerator]


class RenderOutput(TypedDict):
    widgets: List[Dict]
    has_errors: bool


class TemplateRenderer:
    def __init__(self, template: Template) -> None:
        self._raw_template = template

    @property
    def template(self) -> Template:
        # Avoids using mutable objects as widgets currently have a internal state
        return copy.deepcopy(self._raw_template)

    def parse_state(self, raw_state: Dict) -> State:
        parsed = {}
        for idx, widget in enumerate(self.template):
            if isinstance(widget, InputWidget) and widget.key(idx) in raw_state:
                parsed[widget.key(idx)] = widget.parse_value(
                    raw_state.get(widget.key(idx))
                )
        return State(parsed)

    def render(self, state: State) -> RenderOutput:
        rendered_widgets = []
        has_errors = False
        for idx, widget in enumerate(self.template):
            if isinstance(widget, InputWidget):
                value = state.get(widget.key(idx))
                if value is not None:
                    widget.value = value
                if widget.key(idx) in state:
                    widget.set_errors()
                if widget.has_errors() > 0:
                    has_errors = True

            rendered = widget.render()
            rendered_widgets.append(rendered)

        return RenderOutput(widgets=rendered_widgets, has_errors=has_errors)



import inspect
import types
from typing import Dict, List, Optional, Union, cast, overload

from abstra_internals.controllers.sdk_context import SDKContextStore
from abstra_internals.controllers.sdk_forms import FormSDKController
from abstra_internals.entities.forms.form_state import State
from abstra_internals.entities.forms.steps import (
    ComputationStep,
    GeneratorStep,
    PageStep,
    Step,
)
from abstra_internals.entities.forms.template import (
    Template,
    TemplateFunction,
    TemplateGeneratorFunction,
)
from abstra_internals.entities.forms.widgets.widget_base import BaseWidget, InputWidget
from abstra_internals.interface.sdk.forms.exceptions import InvalidRunInputError
from abstra_internals.utils.code import always_returns_none

Runnable = Union[
    Step, TemplateFunction, Template, TemplateGeneratorFunction, types.FunctionType
]


class Form:
    controller: FormSDKController
    state: State
    hide_steps: bool

    def _is_widgets_list(self, runnable: Runnable) -> bool:
        return isinstance(runnable, list) and all(
            isinstance(w, BaseWidget) for w in runnable
        )

    def _create_step(self, runnable: Runnable) -> Step:
        if isinstance(runnable, Step):
            return runnable
        elif self._is_widgets_list(runnable):
            print("PageStep list")
            return PageStep(lambda _: cast(List, runnable))
        elif inspect.isgeneratorfunction(runnable):
            return GeneratorStep(runnable)
        elif isinstance(runnable, types.FunctionType):
            if always_returns_none(runnable):
                return ComputationStep(runnable)
            return PageStep(runnable)

        raise InvalidRunInputError(str(type(runnable)))

    def __init__(
        self,
        runnables: List[Runnable],
        *,
        initial_state: Optional[Dict] = None,
        hide_steps: bool = False,
    ):
        self.steps = [self._create_step(runnable) for runnable in runnables]
        self.controller = SDKContextStore.get_by_thread().form_sdk
        self.state = State(initial_state or {})
        self.hide_steps = hide_steps

    def run(self) -> Dict:
        return self.controller.run_form(
            steps=self.steps, state=self.state, hide_steps=self.hide_steps
        )


@overload
def run(
    runnables: List[Runnable],
    state: Optional[Dict] = None,
    hide_steps: bool = False,
) -> State: ...


@overload
def run(
    runnables: BaseWidget,
    state: Optional[Dict] = None,
    hide_steps: bool = False,
) -> Optional[object]: ...


def run(
    runnables: Union[List[Runnable], BaseWidget],
    state: Optional[Dict] = None,
    hide_steps: bool = False,
) -> Union[State, None, object]:
    # Single widget run
    if isinstance(runnables, BaseWidget):
        state = Form([[runnables]], initial_state=state, hide_steps=hide_steps).run()
        if isinstance(runnables, InputWidget):
            return state[runnables._key]
        return None

    return Form(runnables, initial_state=state, hide_steps=hide_steps).run()



import datetime
import re
import time
from typing import List, Optional, Union

from abstra_internals.entities.forms.widgets.widget_base import InputWidget


class DateInput(InputWidget):
    type = "date-input"
    value: Optional[datetime.date]

    def __init__(
        self,
        label: str,
        *,
        key: Optional[str] = None,
        required: bool = True,
        hint: Optional[str] = None,
        full_width: bool = False,
        disabled: bool = False,
        value: Optional[datetime.date] = None,
        errors: Optional[Union[List[str], str]] = None,
    ):
        self.label = label
        self._key = key or label
        self.required = required
        self.hint = hint
        self.full_width = full_width
        self.disabled = disabled
        self.value = value
        self.errors = self._init_errors(errors)

    def render(self):
        return {
            "type": self.type,
            "key": self._key,
            "hint": self.hint,
            "label": self.label,
            "value": self.serialize_value(),
            "required": self.required,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
            "errors": self.errors,
        }

    def serialize_value(self) -> str:
        if isinstance(self.value, datetime.date):
            date = self.value.isoformat()
            if "T" in date:
                return date.split("T")[0]
            return date
        if isinstance(self.value, time.struct_time):
            return (
                datetime.datetime.fromtimestamp(time.mktime(self.value))
                .date()
                .isoformat()
            )
        if isinstance(self.value, str) and re.match(
            "^\\d{4}-\\d{2}-\\d{2}$", self.value
        ):
            return self.value
        return ""

    def parse_value(self, value: str) -> Optional[datetime.date]:
        if not value:
            return None
        try:
            split_value = value.split("T")[0].split("-")
            year = int(split_value[0])
            month = int(split_value[1])
            day = int(split_value[2])
            return datetime.date(year, month, day)
        except Exception:
            raise ValueError("Invalid date string format. Expected YYYY-MM-DD.")



from typing import List, Optional, Union

from abstra_internals.entities.forms.widgets.widget_base import InputWidget


class ToggleInput(InputWidget):
    type = "toggle-input"
    value: bool

    def __init__(
        self,
        label: str,
        *,
        key: Optional[str] = None,
        on_text: str = "Yes",
        off_text: str = "No",
        required: bool = True,
        hint: Optional[str] = None,
        full_width: bool = False,
        disabled: bool = False,
        errors: Optional[Union[List[str], str]] = None,
        value: bool = False,
    ):
        self.label = label
        self._key = key or label
        self.on_text = on_text
        self.off_text = off_text
        self.value = value
        self.required = required
        self.hint = hint
        self.full_width = full_width
        self.disabled = disabled
        self.errors = self._init_errors(errors)

    def is_value_unset(self):
        return False

    def render(self):
        return {
            "type": self.type,
            "key": self._key,
            "label": self.label,
            "onText": self.on_text,
            "offText": self.off_text,
            "value": self.serialize_value(),
            "required": self.required,
            "hint": self.hint,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
            "errors": self.errors,
        }

    def serialize_value(self) -> bool:
        return self.value or False

    def parse_value(self, value: bool) -> bool:
        return bool(value)



from abstra_internals.entities.forms.widgets.widget_base import OutputWidget


class MarkdownOutput(OutputWidget):
    type = "markdown-output"

    def __init__(
        self,
        text: str,
        *,
        full_width: bool = False,
    ):
        self.text = text
        self.full_width = full_width

    def render(self):
        return {"type": self.type, "text": self.text, "fullWidth": self.full_width}



from typing import List, Optional, Union

from abstra_internals.entities.forms.widgets.widget_base import InputWidget


class RatingInput(InputWidget):
    type = "rating-input"
    value: Optional[float]

    def __init__(
        self,
        label: str,
        *,
        key: Optional[str] = None,
        value: Optional[float] = None,
        required: bool = True,
        hint: Optional[str] = None,
        full_width: bool = False,
        max: Optional[int] = None,
        char: str = "⭐️",
        disabled: bool = False,
        errors: Optional[Union[List[str], str]] = None,
    ):
        self.label = label
        self._key = key or label
        self.value = value
        self.required = required
        self.hint = hint
        self.full_width = full_width
        self.max = max
        self.char = char
        self.disabled = disabled
        self.errors = self._init_errors(errors)

    def render(self):
        return {
            "type": self.type,
            "key": self._key,
            "label": self.label,
            "value": self.value,
            "required": self.required,
            "hint": self.hint,
            "fullWidth": self.full_width,
            "max": self.max,
            "char": self.char,
            "disabled": self.disabled,
            "errors": self.errors,
        }



from datetime import datetime
from typing import List, Optional, Tuple, Union

from abstra_internals.entities.forms.widgets.response_types import AppointmentSlot
from abstra_internals.entities.forms.widgets.widget_base import InputWidget


class AppointmentInput(InputWidget):
    type = "appointment-input"
    slots: List[AppointmentSlot]

    def __init__(
        self,
        label: str,
        *,
        key: Optional[str] = None,
        required: bool = True,
        hint: Optional[str] = None,
        full_width: bool = False,
        disabled: bool = False,
        slots: Optional[List[AppointmentSlot]] = None,
        errors: Optional[Union[List[str], str]] = None,
        value: Optional[Union[AppointmentSlot, dict, Tuple[datetime, datetime]]] = None,
    ) -> None:
        self.label = label
        self._key = key or label
        self.required = required
        self.hint = hint
        self.full_width = full_width
        self.disabled = disabled
        self.slots = AppointmentSlot.from_list(slots or [])
        self.errors = self._init_errors(errors)
        self.value = value

    @property
    def value(self) -> Optional[AppointmentSlot]:
        return self._value

    @value.setter
    def value(
        self, value: Optional[Union[AppointmentSlot, dict, Tuple[datetime, datetime]]]
    ) -> None:
        if value is None:
            self._value = None
        if isinstance(value, AppointmentSlot):
            self._value = value
        elif isinstance(value, dict):
            self._value = AppointmentSlot.from_dict(value)
        elif isinstance(value, tuple):
            self._value = AppointmentSlot(*value)
        else:
            raise ValueError(f"Invalid value type {type(value)}")

    def render(self):
        return {
            "type": self.type,
            "key": self._key,
            "label": self.label,
            "hint": self.hint,
            "value": self.serialize_value(),
            "required": self.required,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
            "slots": [s.to_dict() for s in self.slots],
            "errors": self.errors,
        }

    def serialize_value(self) -> Optional[int]:
        return self.slots.index(self.value) if self.value is not None else None

    def parse_value(self, value: Optional[int]) -> Optional[AppointmentSlot]:
        if value is None:
            return None
        return self.slots[value]



from abstra_internals.entities.forms.widgets.widget_base import OutputWidget


class HtmlOutput(OutputWidget):
    type = "html-output"

    def __init__(
        self,
        html: str,
        *,
        full_width: bool = False,
    ):
        self.html = html
        self.full_width = full_width

    def render(self):
        return {"type": self.type, "html": self.html, "fullWidth": self.full_width}



from typing import List, Optional, TypedDict, Union

from abstra_internals.entities.forms.widgets.file_upload import upload_widget_file
from abstra_internals.entities.forms.widgets.widget_base import (
    InputWidget,
    MultipleHandler,
)


class CardOption(TypedDict):
    title: Optional[str]
    subtitle: Optional[str]
    image: Optional[str]
    description: Optional[str]
    topLeftExtra: Optional[str]
    topRightExtra: Optional[str]


class CardsInput(InputWidget):
    type = "cards-input"
    value: Union[List[CardOption], CardOption, None]
    multiple: bool = False
    multiple_handler: MultipleHandler

    def __init__(
        self,
        label: str,
        options: List[CardOption],
        *,
        key: Optional[str] = None,
        searchable: bool = False,
        required: bool = True,
        hint: Optional[str] = None,
        columns: Optional[int] = 2,
        full_width: bool = False,
        layout: Optional[str] = "list",
        disabled: bool = False,
        min: Optional[int] = None,
        max: Optional[int] = None,
        errors: Optional[Union[List[str], str]] = None,
        value: Optional[Union[List[CardOption], CardOption]] = None,
    ):
        self.label = label
        self._key = key or label
        self.options = [
            {**opt, "image": upload_widget_file(opt.get("image") or "")}
            for opt in options
        ]
        self.searchable = searchable
        self.required = required
        self.hint = hint
        self.columns = columns
        self.full_width = full_width
        self.layout = layout
        self.disabled = disabled
        self.multiple = False
        self.empty_value = []
        self.min = min
        self.max = max
        self.multiple_handler = MultipleHandler(
            self.multiple, self.min, self.max, self.required
        )
        self.errors = self._init_errors(errors)
        self.value = value

    def render(self):
        return {
            "type": self.type,
            "key": self._key,
            "label": self.label,
            "hint": self.hint,
            "options": self.options,
            "multiple": self.multiple,
            "searchable": self.searchable,
            "required": self.required,
            "columns": self.columns,
            "fullWidth": self.full_width,
            "layout": self.layout,
            "disabled": self.disabled,
            "errors": self.errors,
            "min": self.min,
            "max": self.max,
            "value": self.serialize_value(),
        }

    def _run_validators(self) -> List[str]:
        return self.multiple_handler.validate(self.value)

    def serialize_value(self) -> List:
        return self.multiple_handler.value_to_list(self.value)

    def parse_value(self, value):
        return self.multiple_handler.value_to_list_or_value(value)



import datetime
import pkgutil
import webbrowser
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import flask

from abstra_internals.cloud_api import get_api_key_info, get_project_info
from abstra_internals.controllers.linters import check_linters
from abstra_internals.credentials import (
    delete_credentials,
    get_credentials,
    resolve_headers,
    set_credentials,
)
from abstra_internals.interface.cli.deploy import deploy
from abstra_internals.repositories.email import EmailRepository
from abstra_internals.repositories.execution import ExecutionFilter, ExecutionRepository
from abstra_internals.repositories.execution_logs import (
    ExecutionLogsRepository,
    LogEntry,
)
from abstra_internals.repositories.factory import Repositories
from abstra_internals.repositories.jwt_signer import EditorJWTRepository, JWTRepository
from abstra_internals.repositories.keyvalue import KVRepository
from abstra_internals.repositories.producer import ProducerRepository
from abstra_internals.repositories.project.project import (
    FormStage,
    HookStage,
    JobStage,
    ProjectRepository,
    ScriptStage,
    Stage,
    StageWithFile,
    StyleSettingsWithSidebar,
)
from abstra_internals.repositories.roles import RolesRepository
from abstra_internals.repositories.tasks import TasksRepository
from abstra_internals.repositories.users import UsersRepository
from abstra_internals.services.requirements import RequirementsRepository
from abstra_internals.settings import Settings
from abstra_internals.templates import (
    ensure_abstraignore,
    ensure_gitignore,
    new_form_code,
    new_hook_code,
    new_job_code,
    new_script_code,
)
from abstra_internals.utils.dot_abstra import TEST_DATA_FILE
from abstra_internals.utils.file import files_from_directory, module2path, path2module
from abstra_internals.utils.validate import validate_json


class UnknownNodeTypeError(Exception):
    def __init__(self, node_type: str):
        self.node_type = node_type

    def __str__(self):
        return f"Unknown node type {self.node_type}"


class SelfTransitionError(Exception):
    def __init__(self, node_type: str, node_id: str):
        self.node_type = node_type
        self.node_id = node_id

    def __str__(self):
        return "You can't add a transition to itself."


class TransitionToJobError(Exception):
    def __init__(
        self, source_type: str, source_id: str, target_type: str, target_id: str
    ):
        self.source_type = source_type
        self.source_id = source_id
        self.target_type = target_type
        self.target_id = target_id

    def __str__(self):
        return "You can't add a transition to a job. Use a script instead."


class DoubleTransitionError(Exception):
    def __init__(
        self, source_type: str, source_id: str, target_type: str, target_id: str
    ):
        self.source_type = source_type
        self.source_id = source_id
        self.target_type = target_type
        self.target_id = target_id

    def __str__(self):
        return "You can't add the same transition twice."


class MainController:
    kv_repository: KVRepository
    jwt_repository: JWTRepository
    email_repository: EmailRepository
    tasks_repository: TasksRepository
    users_repository: UsersRepository
    roles_repository: RolesRepository
    producer_repository: ProducerRepository
    execution_repository: ExecutionRepository
    web_editor_repository: EditorJWTRepository
    execution_logs_repository: ExecutionLogsRepository

    def __init__(self, repositories: Repositories):
        ProjectRepository.initialize_or_migrate()

        RequirementsRepository.ensure("abstra")
        ensure_abstraignore(Settings.root_path)
        ensure_gitignore(Settings.root_path)

        self.repositories = repositories

        self.kv_repository = repositories.kv
        self.jwt_repository = repositories.jwt
        self.email_repository = repositories.email
        self.users_repository = repositories.users
        self.roles_repository = repositories.roles
        self.tasks_repository = repositories.tasks
        self.producer_repository = repositories.producer
        self.execution_repository = repositories.execution
        self.web_editor_repository = repositories.editor_jwt
        self.execution_logs_repository = repositories.execution_logs

    def deploy(self):
        rules = check_linters()

        issues = []
        for rule in rules:
            if rule.type in ["error", "security", "bug"]:
                issues += rule.issues

        if len(issues) > 0:
            raise Exception(
                "Please fix all linter issues before deploying your project."
            )

        deploy()

    def reset_repositories(self):
        self.execution_repository.clear()
        self.tasks_repository.clear()

    def get_workspace(self) -> StyleSettingsWithSidebar:
        project = ProjectRepository.load()
        return project.get_workspace()

    def get_stage(self, id: str) -> Optional[Stage]:
        project = ProjectRepository.load()
        return project.get_stage(id)

    def get_action(self, id: str) -> Optional[Stage]:
        project = ProjectRepository.load()
        return project.get_action(id)

    def get_async_stage_ids(self):
        project = ProjectRepository.load()
        job_ids = [stage.id for stage in project.jobs]
        script_ids = [stage.id for stage in project.scripts]
        return job_ids + script_ids

    def __ensure_case(self, path: str):
        file_dirs = [p for p in Settings.root_path.iterdir()]
        if path in file_dirs:
            return

        conflicting_paths = [p for p in file_dirs if p.name.lower() == path.lower()]
        if len(conflicting_paths) == 1:
            conflicting_paths[0].rename(Settings.root_path.joinpath(path))
            return

        raise Exception(
            f"File {path} already exists with different casing. Conflict with files ({', '.join(p.name for p in conflicting_paths)})"
        )

    def init_code_file(self, path: str, code: str):
        if Settings.root_path.joinpath(path).is_file():
            self.__ensure_case(path)
            return
        Settings.root_path.joinpath(path).write_text(code, encoding="utf-8")

    def open_file(self, file_path: str, mode: str, create_if_not_exists: bool = False):
        if mode == "module" or mode == "package":
            file_path = str(module2path(file_path, mode == "package"))
        complete_file_path = Settings.root_path.joinpath(file_path)

        if create_if_not_exists and not complete_file_path.is_file():
            complete_file_path.touch()

        webbrowser.open(complete_file_path.absolute().as_uri())

    def read_file(self, file: str):
        file_path = Settings.root_path.joinpath(file)
        if not file_path.is_file():
            return None
        return Settings.root_path.joinpath(file).read_text(encoding="utf-8")

    def check_file(self, file_path: str):
        return Settings.root_path.joinpath(file_path).is_file()

    def check_files(self, file_paths: List[str]):
        return {file_path: self.check_file(file_path) for file_path in file_paths}

    def list_files(self, path: str = ".", mode: str = "file"):
        parent_path = Settings.root_path.joinpath(path)
        if mode in ["file", "image", "python-file"]:
            if mode == "image":
                allowed_suffixes = [
                    ".png",
                    ".jpg",
                    ".jpeg",
                    ".gif",
                    ".svg",
                    ".webp",
                    ".jfif",
                    ".pjp",
                    ".pjpeg",
                ]
            elif mode == "python-file":
                allowed_suffixes = [".py"]
            else:
                allowed_suffixes = None

            return [
                dict(
                    name=str(file.relative_to(parent_path)),
                    path=str(file.relative_to(Settings.root_path)),
                    type="file" if file.is_file() else "dir",
                )
                for file in files_from_directory(parent_path)
                if file.is_dir()
                or not allowed_suffixes
                or file.suffix in allowed_suffixes
            ]

        elif mode == "module":
            return [
                dict(
                    name=name,
                    path=path2module(Path(path).joinpath(name)),
                    type="package" if ispkg else "module",
                )
                for module_finder, name, ispkg in pkgutil.iter_modules(
                    [str(parent_path)]
                )
            ]

    def update_workspace(self, changes: Dict[str, Any]):
        project = ProjectRepository.load()
        project.workspace.update(changes)
        ProjectRepository.save(project)
        return project.workspace

    def is_initial(self, id: str):
        project = ProjectRepository.load()
        stage = project.get_action(id)
        if not stage:
            raise Exception(f"Stage {id} not found")
        return project.is_initial(stage)

    def create_script(
        self, title: str, file: str, workflow_position: Tuple[int, int] = (0, 0)
    ) -> ScriptStage:
        project = ProjectRepository.load()
        script = ScriptStage.create(title, file, workflow_position=workflow_position)
        self.init_code_file(script.file, new_script_code)
        project.add_stage(script)
        ProjectRepository.save(project)

        return script

    def get_scripts(self) -> List[ScriptStage]:
        project = ProjectRepository.load()
        return project.scripts

    def get_script(self, id: str) -> Optional[ScriptStage]:
        project = ProjectRepository.load()
        return project.get_script(id)

    def delete_script(self, id: str, remove_file: bool = False):
        project = ProjectRepository.load()
        project.delete_stage(id, remove_file)
        ProjectRepository.save(project)

    def create_form(
        self, title: str, file: str, workflow_position: Tuple[int, int] = (0, 0)
    ) -> FormStage:
        project = ProjectRepository.load()
        form = FormStage.create(title, file, workflow_position=workflow_position)
        self.init_code_file(form.file, new_form_code)
        project.add_stage(form)
        ProjectRepository.save(project)
        return form

    def get_forms(self) -> List[FormStage]:
        project = ProjectRepository.load()
        return project.forms

    def get_form(self, id: str) -> Optional[FormStage]:
        project = ProjectRepository.load()
        return project.get_form(id)

    def get_form_by_path(self, path: str) -> Optional[FormStage]:
        project = ProjectRepository.load()
        return project.get_form_by_path(path)

    def write_test_data(self, data: str) -> None:
        if not validate_json(data):
            raise Exception("Invalid JSON")
        test_file = Settings.root_path / TEST_DATA_FILE
        test_file.write_text(data, encoding="utf-8")

    def read_test_data(self) -> str:
        test_file = Settings.root_path / TEST_DATA_FILE
        if not test_file.is_file():
            return "{}"
        return test_file.read_text(encoding="utf-8")

    def delete_form(self, id: str, remove_file: bool = False):
        project = ProjectRepository.load()
        project.delete_stage(id, remove_file)
        ProjectRepository.save(project)

    def create_hook(
        self, title: str, file: str, workflow_position: Tuple[int, int] = (0, 0)
    ) -> HookStage:
        project = ProjectRepository.load()
        hook = HookStage.create(title, file, workflow_position=workflow_position)
        self.init_code_file(hook.file, new_hook_code)
        project.add_stage(hook)
        ProjectRepository.save(project)
        return hook

    def get_hook(self, id: str) -> Optional[HookStage]:
        project = ProjectRepository.load()
        return project.get_hook(id)

    def get_hooks(self) -> List[HookStage]:
        project = ProjectRepository.load()
        return project.hooks

    def get_hook_by_path(self, path: str) -> Optional[HookStage]:
        project = ProjectRepository.load()
        return project.get_hook_by_path(path)

    def delete_hook(self, id: str, remove_file: bool = False) -> None:
        project = ProjectRepository.load()
        project.delete_stage(id, remove_file)
        ProjectRepository.save(project)

    def get_jobs(self) -> List[JobStage]:
        project = ProjectRepository.load()
        return project.jobs

    def get_job(self, id: str) -> Optional[JobStage]:
        project = ProjectRepository.load()
        stage = project.get_stage(id)

        if isinstance(stage, JobStage):
            return stage

        return None

    def create_job(
        self, title: str, file: str, workflow_position: Tuple[int, int] = (0, 0)
    ) -> JobStage:
        project = ProjectRepository.load()
        job = JobStage.create(title, file, workflow_position=workflow_position)
        self.init_code_file(job.file, new_job_code)
        project.add_stage(job)
        ProjectRepository.save(project)
        return job

    def update_stage(self, id: str, changes: Dict[str, Any]) -> Stage:
        project = ProjectRepository.load()
        stage = project.get_action(id)

        if not stage:
            raise Exception(f"Stage with id {id} not found")

        if isinstance(stage, StageWithFile) and (
            code_content := changes.pop("code_content", None)
        ):
            Settings.root_path.joinpath(stage.file_path).write_text(
                code_content, encoding="utf-8"
            )

        if test_data := changes.pop("test_data", None):
            self.write_test_data(test_data)

        stage = project.update_stage(stage, changes)
        ProjectRepository.save(project)
        return stage

    def delete_job(self, id: str, remove_file: bool = False):
        project = ProjectRepository.load()
        project.delete_stage(id, remove_file)
        ProjectRepository.save(project)

    def get_stages(self) -> List[Stage]:
        project = ProjectRepository.load()
        return project.workflow_stages

    # Login
    def get_credentials(self):
        return get_credentials()

    def get_login(self):
        headers = resolve_headers()
        if not headers:
            return {"logged": False, "reason": "NO_API_TOKEN"}
        return get_api_key_info(headers)

    def create_login(self, token):
        set_credentials(token)
        return self.get_login()

    def delete_login(self):
        delete_credentials()
        return self.get_login()

    # Project
    def get_project_info(self):
        headers = resolve_headers()
        if headers is None:
            flask.abort(401)
        return get_project_info(headers)

    # access_control
    def list_access_controls(self):
        project = ProjectRepository.load()
        return [s.to_access_dto() for s in project.secured_stages]

    def update_access_controls(self, changes: List[Dict[str, Any]]):
        project = ProjectRepository.load()
        response = project.update_access_controls(changes)
        ProjectRepository.save(project)
        return response

    def get_access_control_by_stage_id(self, id):
        project = ProjectRepository.load()
        return project.get_access_control_by_stage_id(id)

    # logs
    def get_executions(self, filter: ExecutionFilter):
        return self.execution_repository.list(filter)

    def get_logs(self, id: str):
        return self.execution_logs_repository.get(id)

    # Worker lifecycle

    def child_exit(self, *, app_id: str, worker_id: str, err_msg: str):
        killed_executions = self.execution_repository.find_by_worker(
            worker_id=worker_id,
            status="running",
            app_id=app_id,
        )

        # update executions
        for execution in killed_executions:
            # Add log entry
            err_log = LogEntry(
                execution_id=execution.id,
                created_at=datetime.datetime.now(),
                payload={"text": err_msg},
                sequence=999999,
                event="stderr",
            )
            self.execution_logs_repository.save(err_log)

            # Update execution status
            self.execution_repository.set_failure_by_id(execution_id=execution.id)

            self.tasks_repository.set_locked_tasks_to_pending(execution.id)

    def self_exit(self, *, app_id: str, err_msg: str):
        exited_execs = self.execution_repository.find_by_app(
            status="running",
            app_id=app_id,
        )

        # update executions
        for execution in exited_execs:
            # Add log entry
            err_log = LogEntry(
                execution_id=execution.id,
                created_at=datetime.datetime.now(),
                payload={"text": err_msg},
                sequence=999999,
                event="stderr",
            )
            self.execution_logs_repository.save(err_log)

            # Update execution status
            self.execution_repository.set_failure_by_id(execution_id=execution.id)

            self.tasks_repository.set_locked_tasks_to_pending(execution.id)



from typing import List, Optional, Union

from abstra_internals.entities.forms.widgets.widget_base import (
    AbstraOption,
    InputWidget,
    MultipleHandler,
    OptionsHandler,
)


class DropdownInput(InputWidget):
    type = "dropdown-input"
    multiple_handler: MultipleHandler
    options_handler: OptionsHandler
    value: Union[List[object], object]

    def __init__(
        self,
        label: str,
        options: List[AbstraOption],
        *,
        key: Optional[str] = None,
        required: bool = True,
        hint: Optional[str] = None,
        placeholder: str = "",
        full_width: bool = False,
        disabled: bool = False,
        multiple: bool = False,
        min: Optional[int] = None,
        max: Optional[int] = None,
        errors: Optional[Union[List[object], object]] = None,
        value: Union[List[str], None] = None,
    ):
        self.label = label
        self._key = key or label
        self.options = options
        self.required = required
        self.hint = hint
        self.placeholder = placeholder
        self.full_width = full_width
        self.disabled = disabled
        self.multiple = multiple
        self.min = min
        self.max = max
        self.value = value
        self.multiple_handler = MultipleHandler(
            self.multiple, self.min, self.max, self.required
        )
        self.options_handler = OptionsHandler(self.options)
        self.errors = self._init_errors(errors)

    def render(self):
        return {
            "type": self.type,
            "key": self._key,
            "label": self.label,
            "options": self.options_handler.serialized_options(),
            "hint": self.hint,
            "multiple": self.multiple,
            "placeholder": self.placeholder,
            "value": self.serialize_value(),
            "required": self.required,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
            "errors": self.errors,
            "min": self.min,
            "max": self.max,
        }

    def _run_validators(self) -> List[str]:
        return self.multiple_handler.validate(self.value)

    def serialize_value(self) -> List:
        list_values = self.multiple_handler.value_to_list(self.value)
        return self.options_handler.ids_from_values(list_values)

    def parse_value(self, value: Union[List[str], None]):
        list_values = self.options_handler.values_from_ids(value)
        return self.multiple_handler.value_to_list_or_value(list_values)



from typing import List, Optional, Union

from abstra_internals.entities.forms.widgets.widget_base import (
    InputWidget,
    NumberValueHandler,
)


class NumberInput(InputWidget):
    type = "number-input"
    value: Union[float, None]

    def __init__(
        self,
        label: str,
        *,
        key: Optional[str] = None,
        placeholder: Optional[str] = None,
        required: bool = True,
        hint: Optional[str] = None,
        full_width: bool = False,
        disabled: bool = False,
        initial_value: Optional[float] = None,
        min: Optional[float] = None,
        max: Optional[float] = None,
        errors: Optional[Union[List[str], str]] = None,
        value: Optional[float] = None,
    ):
        self.label = label
        self._key = key or label
        self.placeholder = placeholder
        self.required = required
        self.hint = hint
        self.full_width = full_width
        self.disabled = disabled
        self.value = initial_value
        self.min = min
        self.max = max
        self.errors = self._init_errors(errors)
        self.number_value_handler = NumberValueHandler(
            min=self.min, max=self.max, required=self.required
        )
        self.value = value

    def render(self):
        return {
            "type": self.type,
            "key": self._key,
            "label": self.label,
            "value": self.value,
            "placeholder": self.placeholder,
            "required": self.required,
            "hint": self.hint,
            "fullWidth": self.full_width,
            "min": self.min,
            "max": self.max,
            "disabled": self.disabled,
            "errors": self.errors,
        }

    def _run_validators(self) -> List[str]:
        return self.number_value_handler.validate(self.value)
